@@ -109,7 +109,7 @@ bool DwarfExpression::addMachineReg(const TargetRegisterInfo &TRI,
       return true;
     }
     // Try getting dwarf register for virtual register anyway, eg. for NVPTX.
-    int64_t Reg = TRI.getDwarfRegNum(MachineReg, false);
+    int64_t Reg = TRI.getDwarfRegNumForVirtReg(MachineReg, false);
     if (Reg > 0) {
       DwarfRegs.push_back(Register::createRegister(Reg, nullptr));
       return true;
@@ -157,7 +157,7 @@ bool DwarfExpression::addMachineReg(const TargetRegisterInfo &TRI,
     unsigned Size = TRI.getSubRegIdxSize(Idx);
     unsigned Offset = TRI.getSubRegIdxOffset(Idx);
     Reg = TRI.getDwarfRegNum(SR, false);
-    if (Reg < 0)
+    if (Reg < 0 || Offset + Size > RegSize)
       continue;
 
     // Used to build the intersection between the bits we already
@@ -197,8 +197,17 @@ void DwarfExpression::addStackValue() {
     emitOp(dwarf::DW_OP_stack_value);
 }
 
+void DwarfExpression::addBooleanConstant(int64_t Value) {
+  assert(isImplicitLocation() || isUnknownLocation());
+  LocationKind = Implicit;
+  if (Value == 0)
+    emitOp(dwarf::DW_OP_lit0);
+  else
+    emitOp(dwarf::DW_OP_lit1);
+}
+
 void DwarfExpression::addSignedConstant(int64_t Value) {
-  if (IsPoisonedExpr)
+  if (IsPoisonedExpr || !IsImplemented)
     return;
   assert(isImplicitLocation() || isUnknownLocation());
   LocationKind = Implicit;
@@ -207,7 +216,7 @@ void DwarfExpression::addSignedConstant(int64_t Value) {
 }
 
 void DwarfExpression::addUnsignedConstant(uint64_t Value) {
-  if (IsPoisonedExpr)
+  if (IsPoisonedExpr || !IsImplemented)
     return;
   assert(isImplicitLocation() || isUnknownLocation());
   LocationKind = Implicit;
@@ -215,7 +224,7 @@ void DwarfExpression::addUnsignedConstant(uint64_t Value) {
 }
 
 void DwarfExpression::addUnsignedConstant(const APInt &Value) {
-  if (IsPoisonedExpr)
+  if (IsPoisonedExpr || !IsImplemented)
     return;
   assert(isImplicitLocation() || isUnknownLocation());
   LocationKind = Implicit;
@@ -237,7 +246,7 @@ void DwarfExpression::addUnsignedConstant(const APInt &Value) {
 }
 
 void DwarfExpression::addConstantFP(const APFloat &APF, const AsmPrinter &AP) {
-  if (IsPoisonedExpr)
+  if (IsPoisonedExpr || !IsImplemented)
     return;
   assert(isImplicitLocation() || isUnknownLocation());
   APInt API = APF.bitcastToAPInt();
@@ -269,7 +278,7 @@ bool DwarfExpression::addMachineRegExpression(const TargetRegisterInfo &TRI,
                                               DIExpressionCursor &ExprCursor,
                                               llvm::Register MachineReg,
                                               unsigned FragmentOffsetInBits) {
-  if (IsPoisonedExpr)
+  if (IsPoisonedExpr || !IsImplemented)
     return true;
   auto Fragment = ExprCursor.getFragmentInfo();
   if (!addMachineReg(TRI, MachineReg, Fragment ? Fragment->SizeInBits : ~1U)) {
@@ -526,6 +535,8 @@ bool DwarfExpression::addExpression(
   // and not any other parts of the following DWARF expression.
   assert(!IsEmittingEntryValue && "Can't emit entry value around expression");
 
+  if (!IsImplemented)
+    return false;
   IsPoisonedExpr = false;
 
   std::optional<DIExpression::ExprOperand> PrevConvertOp;
@@ -544,8 +555,7 @@ bool DwarfExpression::addExpression(
 
     switch (OpNum) {
     case dwarf::DW_OP_LLVM_poisoned:
-      emitOp(dwarf::DW_OP_LLVM_user);
-      emitOp(dwarf::DW_OP_LLVM_USER_undefined);
+      emitUserOp(dwarf::DW_OP_LLVM_undefined);
       LocationKind = Unknown;
       break;
     case dwarf::DW_OP_LLVM_arg:
@@ -640,12 +650,15 @@ bool DwarfExpression::addExpression(
     case dwarf::DW_OP_dup:
     case dwarf::DW_OP_push_object_address:
     case dwarf::DW_OP_over:
+    case dwarf::DW_OP_rot:
     case dwarf::DW_OP_eq:
     case dwarf::DW_OP_ne:
     case dwarf::DW_OP_gt:
     case dwarf::DW_OP_ge:
     case dwarf::DW_OP_lt:
     case dwarf::DW_OP_le:
+    case dwarf::DW_OP_neg:
+    case dwarf::DW_OP_abs:
       emitOp(OpNum);
       break;
     case dwarf::DW_OP_deref:
@@ -733,6 +746,8 @@ bool DwarfExpression::addExpression(
 void DwarfExpression::addExpression(DIExpression::NewElementsRef Expr,
                                     ArrayRef<DbgValueLocEntry> ArgLocEntries,
                                     const TargetRegisterInfo *TRI) {
+  if (!IsImplemented)
+    return;
   assert(!IsPoisonedExpr && "poisoned exprs should have old elements");
   this->ArgLocEntries = ArgLocEntries;
   this->TRI = TRI;
@@ -745,13 +760,17 @@ void DwarfExpression::addExpression(DIExpression::NewElementsRef Expr,
     }
   }
   buildAST(Expr);
-  traverse(ASTRoot.get(), ValueKind::LocationDesc);
+  traverse(ASTRoot.get(), ValueKind::LocationDesc,
+           /*PermitDivergentAddrSpace=*/
+           PermitDivergentAddrSpaceResult && !IsFragment);
   if (FragOp)
     addOpPiece(FragOp->getBitSize());
+  if (!IsImplemented)
+    emitUserOp(dwarf::DW_OP_LLVM_undefined);
   IsFragment = false;
   ASTRoot.reset();
   this->TRI = nullptr;
-  this->ArgLocEntries = std::nullopt;
+  this->ArgLocEntries = {};
 }
 
 /// add masking operations to stencil out a subregister.
@@ -764,12 +783,6 @@ void DwarfExpression::maskSubRegister() {
 }
 
 void DwarfExpression::emitUserOp(uint8_t UserOp, const char *Comment) {
-  if (!EmitHeterogeneousDwarfAsUserOps) {
-    auto OptNonUserOp = dwarf::getNonUserOp(UserOp);
-    assert(OptNonUserOp && "non-user op passed to DwarfExpression::emitUserOp");
-    emitOp(*OptNonUserOp, Comment);
-    return;
-  }
   emitOp(dwarf::DW_OP_LLVM_user);
   emitOp(UserOp);
 }
@@ -786,7 +799,7 @@ void DwarfExpression::finalize() {
 }
 
 void DwarfExpression::addFragmentOffset(const DIExpression *Expr) {
-  if (!Expr)
+  if (!Expr || !IsImplemented)
     return;
 
   if (Expr->holdsOldElements() && Expr->isPoisoned())
@@ -842,7 +855,7 @@ void DwarfExpression::emitLegacyZExt(unsigned FromBits) {
 }
 
 void DwarfExpression::addWasmLocation(unsigned Index, uint64_t Offset) {
-  if (IsPoisonedExpr)
+  if (IsPoisonedExpr || !IsImplemented)
     return;
   emitOp(dwarf::DW_OP_WASM_location);
   emitUnsigned(Index == 4/*TI_LOCAL_INDIRECT*/ ? 0/*TI_LOCAL*/ : Index);
@@ -888,7 +901,8 @@ void DwarfExpression::buildAST(DIExpression::NewElementsRef Elements) {
 using NewOpResult = DwarfExpression::OpResult;
 
 std::optional<NewOpResult>
-DwarfExpression::traverse(Node *OpNode, std::optional<ValueKind> ReqVK) {
+DwarfExpression::traverse(Node *OpNode, std::optional<ValueKind> ReqVK,
+                          bool PermitDivergentAddrSpace) {
   std::optional<NewOpResult> Result =
       std::visit([&](auto &&E) { return traverse(E, OpNode->getChildren()); },
                  OpNode->getElement());
@@ -896,6 +910,12 @@ DwarfExpression::traverse(Node *OpNode, std::optional<ValueKind> ReqVK) {
     IsImplemented = false;
     return Result;
   }
+  if (Result->DivergentAddrSpace && !PermitDivergentAddrSpace) {
+    // FIXME: When DWARF supports address space conversions, generate a
+    // DW_OP_convert here to convert to the required address space.
+    IsImplemented = false;
+    return Result;
+  }
   OpNode->setIsLowered();
   OpNode->setResultType(Result->Ty);
   return ReqVK ? convertValueKind(*Result, *ReqVK) : Result;
@@ -905,12 +925,12 @@ NewOpResult DwarfExpression::convertValueKind(const NewOpResult &Res,
                                               ValueKind ReqVK) {
   if (Res.VK == ValueKind::Value && ReqVK == ValueKind::LocationDesc) {
     emitOp(dwarf::DW_OP_stack_value);
-    return {Res.Ty, ValueKind::LocationDesc};
+    return {Res.Ty, ValueKind::LocationDesc, Res.DivergentAddrSpace};
   }
 
   if (Res.VK == ValueKind::LocationDesc && ReqVK == ValueKind::Value) {
     readToValue(Res.Ty);
-    return {Res.Ty, ValueKind::Value};
+    return {Res.Ty, ValueKind::Value, Res.DivergentAddrSpace};
   }
 
   return Res;
@@ -951,7 +971,7 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Arg Arg,
 
   if (Entry.isLocation()) {
     assert(DwarfRegs.empty() && "unconsumed registers?");
-    if (!addMachineReg(*TRI, Entry.getLoc().getReg())) {
+    if (!TRI || !addMachineReg(*TRI, Entry.getLoc().getReg())) {
       DwarfRegs.clear();
       return std::nullopt;
     }
@@ -968,11 +988,22 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Arg Arg,
     SubRegOffset /= 8;
     SubRegSize /= 8;
 
+    auto focusThreadIfRequired = [this](int64_t DwarfRegNo) {
+      // FIXME: This should be represented in the DIExpression.
+      if (auto LaneSize = TRI->getDwarfRegLaneSize(DwarfRegNo, false)) {
+        emitUserOp(dwarf::DW_OP_LLVM_push_lane);
+        emitConstu(*LaneSize);
+        emitOp(dwarf::DW_OP_mul);
+        emitUserOp(dwarf::DW_OP_LLVM_offset);
+      }
+    };
+
     if (Regs.size() == 1) {
       addReg(Regs[0].DwarfRegNo, Regs[0].Comment);
+      focusThreadIfRequired(Regs[0].DwarfRegNo);
 
       if (SubRegOffset) {
-        emitUserOp(dwarf::DW_OP_LLVM_USER_offset_uconst);
+        emitUserOp(dwarf::DW_OP_LLVM_offset_uconst);
         emitUnsigned(SubRegOffset);
       }
 
@@ -994,12 +1025,14 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Arg Arg,
     for (auto &Reg : Regs) {
       if (Reg.SubRegSize % 8)
         return std::nullopt;
-      if (Reg.DwarfRegNo >= 0)
+      if (Reg.DwarfRegNo >= 0) {
         addReg(Reg.DwarfRegNo, Reg.Comment);
+        focusThreadIfRequired(Regs[0].DwarfRegNo);
+      }
       emitOp(dwarf::DW_OP_piece);
       emitUnsigned(Reg.SubRegSize / 8);
     }
-    emitUserOp(dwarf::DW_OP_LLVM_USER_piece_end);
+    emitUserOp(dwarf::DW_OP_LLVM_piece_end);
 
     if (IsFragment) {
       emitOp(dwarf::DW_OP_swap);
@@ -1075,14 +1108,28 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::AddrOf AddrOf,
 
 std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Convert Convert,
                                                      ChildrenT Children) {
-  auto Child = traverse(Children[0].get(), ValueKind::Value);
+  auto Child = traverse(Children[0].get(), /*RequiredVK=*/std::nullopt,
+                        /*PermitDivergentAddrSpace=*/true);
   if (!Child)
     return std::nullopt;
 
   Type *DestTy = Convert.getResultType();
+  if (Child->Ty->isPointerTy() && DestTy->isPointerTy() &&
+      Child->Ty->getPointerAddressSpace() != DestTy->getPointerAddressSpace()) {
+    unsigned DivAddrSpace = Child->DivergentAddrSpace
+                                ? *Child->DivergentAddrSpace
+                                : Child->Ty->getPointerAddressSpace();
+    return NewOpResult{DestTy, Child->VK, DivAddrSpace};
+  }
+
   if (!Child->Ty->isIntegerTy() || !DestTy->isIntegerTy())
     return std::nullopt;
 
+  // If we're not dealing with the divergent address space case, Convert
+  // requires a value operand.
+  if (Child->VK == ValueKind::LocationDesc)
+    readToValue(Child->Ty);
+
   uint64_t ToBits = DestTy->getPrimitiveSizeInBits().getFixedValue();
   uint64_t FromBits = Child->Ty->getPrimitiveSizeInBits().getFixedValue();
 
@@ -1118,7 +1165,8 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::SExt SExt,
 
 std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Deref Deref,
                                                      ChildrenT Children) {
-  auto Child = traverse(Children[0].get(), ValueKind::LocationDesc);
+  auto Child = traverse(Children[0].get(), ValueKind::LocationDesc,
+                        /*PermitDivergentAddrSpace=*/true);
   if (!Child)
     return std::nullopt;
 
@@ -1129,13 +1177,14 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Deref Deref,
   PointerType *PointerResultType = dyn_cast<PointerType>(Child->Ty);
   assert(PointerResultType && "Expected PointerType, but got something else");
 
-  uint64_t PointerSizeInBits = AP.getDataLayout().getPointerSizeInBits(
-      PointerResultType->getAddressSpace());
+  unsigned PointerLLVMAddrSpace = Child->DivergentAddrSpace
+                                      ? *Child->DivergentAddrSpace
+                                      : PointerResultType->getAddressSpace();
+  uint64_t PointerSizeInBits =
+      AP.getDataLayout().getPointerSizeInBits(PointerLLVMAddrSpace);
   assert(PointerSizeInBits % 8 == 0 && "Expected multiple of 8");
 
   uint64_t PointerSizeInBytes = PointerSizeInBits / 8;
-
-  unsigned PointerLLVMAddrSpace = PointerResultType->getAddressSpace();
   auto PointerDWARFAddrSpace = AP.TM.mapToDWARFAddrSpace(PointerLLVMAddrSpace);
   if (!PointerDWARFAddrSpace) {
     LLVM_DEBUG(dbgs() << "Failed to lower DIOpDeref of pointer to addrspace("
@@ -1147,7 +1196,7 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Deref Deref,
   emitOp(dwarf::DW_OP_deref_size);
   emitData1(PointerSizeInBytes);
   emitConstu(*PointerDWARFAddrSpace);
-  emitUserOp(dwarf::DW_OP_LLVM_USER_form_aspace_address);
+  emitUserOp(dwarf::DW_OP_LLVM_form_aspace_address);
 
   // FIXME(KZHURAVL): Is the following result type correct?
   return NewOpResult{Deref.getResultType(), ValueKind::LocationDesc};
@@ -1169,10 +1218,12 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Read Read,
 
 std::optional<NewOpResult>
 DwarfExpression::traverse(DIOp::Reinterpret Reinterpret, ChildrenT Children) {
-  auto Child = traverse(Children[0].get(), ValueKind::LocationDesc);
+  auto Child = traverse(Children[0].get(), ValueKind::LocationDesc,
+                        /*PermitDivergentAddrSpace=*/true);
   if (!Child)
     return Child;
-  return NewOpResult{Reinterpret.getResultType(), Child->VK};
+  return NewOpResult{Reinterpret.getResultType(), Child->VK,
+                     Child->DivergentAddrSpace};
 }
 
 std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Select Select,
@@ -1182,11 +1233,31 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Select Select,
 
 std::optional<NewOpResult> DwarfExpression::traverse(DIOp::Composite Composite,
                                                      ChildrenT Children) {
-  return std::nullopt;
+  if (IsFragment)
+    emitOp(dwarf::DW_OP_lit0);
+
+  for (auto &Child : Children) {
+    auto R = traverse(Child.get(), std::nullopt);
+    if (!R)
+      return std::nullopt;
+    TypeSize Size = R->Ty->getPrimitiveSizeInBits();
+    if (!Size.isFixed() || Size.getFixedValue() % 8 != 0)
+      return std::nullopt;
+    emitOp(dwarf::DW_OP_piece);
+    emitUnsigned(Size.getFixedValue() / 8);
+  }
+  emitUserOp(dwarf::DW_OP_LLVM_piece_end);
+
+  if (IsFragment) {
+    emitOp(dwarf::DW_OP_swap);
+    emitOp(dwarf::DW_OP_drop);
+  }
+
+  return NewOpResult{Composite.getResultType(), ValueKind::LocationDesc};
 }
 
-std::optional<NewOpResult> DwarfExpression::traverseMathOp(uint8_t DwarfOp,
-                                                           ChildrenT Children) {
+std::optional<NewOpResult>
+DwarfExpression::traverseMathOp(uint8_t DwarfOp, ChildrenT Children) {
   auto LHS = traverse(Children[0].get(), ValueKind::Value);
   if (!LHS)
     return std::nullopt;
@@ -1207,7 +1278,7 @@ DwarfExpression::traverse(DIOp::ByteOffset ByteOffset, ChildrenT Children) {
   if (!RHS)
     return std::nullopt;
 
-  emitUserOp(dwarf::DW_OP_LLVM_USER_offset);
+  emitUserOp(dwarf::DW_OP_LLVM_offset);
   return NewOpResult{ByteOffset.getResultType(), ValueKind::LocationDesc};
 }
 
@@ -1220,7 +1291,7 @@ std::optional<NewOpResult> DwarfExpression::traverse(DIOp::BitOffset BitOffset,
   if (!RHS)
     return std::nullopt;
 
-  emitUserOp(dwarf::DW_OP_LLVM_USER_bit_offset);
+  emitUserOp(dwarf::DW_OP_LLVM_bit_offset);
   return NewOpResult{BitOffset.getResultType(), ValueKind::LocationDesc};
 }
 
@@ -1253,444 +1324,3 @@ void DwarfExpression::readToValue(DwarfExpression::Node *OpNode) {
   assert(OpNode->getResultType() && "Expected non-null result type");
   readToValue(OpNode->getResultType());
 }
-
-void DwarfExprAST::buildDIExprAST() {
-  std::stack<std::unique_ptr<DwarfExprAST::Node>> Operands;
-
-  for (const auto &Op : Lifetime.getLocation()->builder()) {
-    // DIOp::Fragment is not a true operation, and like DW_OP_LLVM_fragment
-    // it can be ignored when interpreting the expression semantically. Just
-    // drop it here during the first walk of the expression and assert it
-    // is not present later.
-    if (std::holds_alternative<DIOp::Fragment>(Op))
-      continue;
-    std::unique_ptr<DwarfExprAST::Node> OpNode =
-        std::make_unique<DwarfExprAST::Node>(Op);
-    size_t OpChildrenCount = DIOp::getNumInputs(OpNode->getElement());
-    if (OpChildrenCount == 0) {
-      Operands.push(std::move(OpNode));
-    } else {
-      for (size_t I = 0; I < OpChildrenCount; ++I) {
-        OpNode->getChildren().insert(
-            OpNode->getChildren().begin(), std::move(Operands.top()));
-        Operands.pop();
-      }
-      Operands.push(std::move(OpNode));
-    }
-  }
-
-  assert(Operands.size() == 1);
-  Root = std::move(Operands.top());
-}
-
-using OpResult = DwarfExprAST::OpResult;
-
-std::optional<OpResult> DwarfExprAST::traverse(Node *OpNode,
-                                               std::optional<ValueKind> ReqVK) {
-  std::optional<OpResult> Result =
-      std::visit([&](auto &&E) { return traverse(E, OpNode->getChildren()); },
-                 OpNode->getElement());
-  if (!Result) {
-    IsImplemented = false;
-    return Result;
-  }
-  OpNode->setIsLowered();
-  OpNode->setResultType(Result->Ty);
-  return ReqVK ? convertValueKind(*Result, *ReqVK) : Result;
-}
-
-OpResult DwarfExprAST::convertValueKind(const OpResult &Res, ValueKind ReqVK) {
-  if (Res.VK == ValueKind::Value && ReqVK == ValueKind::LocationDesc) {
-    emitDwarfOp(dwarf::DW_OP_stack_value);
-    return {Res.Ty, ValueKind::LocationDesc};
-  }
-
-  if (Res.VK == ValueKind::LocationDesc && ReqVK == ValueKind::Value) {
-    readToValue(Res.Ty);
-    return {Res.Ty, ValueKind::Value};
-  }
-
-  return Res;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Arg Arg,
-                                               ChildrenT Children) {
-  uint32_t Index = Arg.getIndex();
-  assert(Index < std::distance(Lifetime.argObjects().begin(),
-                               Lifetime.argObjects().end()));
-  DIObject *ArgObject = *(Lifetime.argObjects().begin() + Index);
-
-  if (!GVFragmentMap)
-    return std::nullopt;
-  auto *Fragment = dyn_cast<DIFragment>(ArgObject);
-  if (!Fragment)
-    return std::nullopt;
-  const auto GVRef = GVFragmentMap->find(Fragment);
-  if (GVRef == GVFragmentMap->end() || !GVRef->getSecond())
-    return std::nullopt;
-  const GlobalVariable *Global = cast<GlobalVariable>(GVRef->getSecond());
-  // FIXME(KZHURAVL): This depends on the target and address space
-  // semantics. For AMDGPU, address space 3 is lds/local/shared.
-  // Need to replace this with a target hook!
-  if (Global->getAddressSpace() == 3) {
-    // Non-generic address space.
-    emitUnsigned(0);
-  } else {
-    const MCSymbol *Sym = AP.getSymbol(Global);
-    DwarfDebug *DD = AP.getDwarfDebug();
-    if (DD->useSplitDwarf()) {
-      bool UseAddrOffsetFormOrExpressions =
-          DD->useAddrOffsetForm() || DD->useAddrOffsetExpressions();
-
-      const MCSymbol *Base = nullptr;
-      if (Sym->isInSection() && UseAddrOffsetFormOrExpressions)
-        Base = DD->getSectionLabel(&Sym->getSection());
-
-      unsigned Index = DD->getAddressPool().getIndex(Base ? Base : Sym);
-      emitDwarfOpAddrx(Index);
-      if (Base && Base != Sym) {
-        emitUnsigned(4);
-        emitDwarfLabelDelta(Sym, Base);
-        emitDwarfOp(dwarf::DW_OP_plus);
-      }
-    } else {
-      CU.getDwarfDebug().addArangeLabel(SymbolCU(&CU, Sym));
-      emitDwarfOp(dwarf::DW_OP_addr);
-      emitDwarfAddr(Sym);
-    }
-  }
-
-  return OpResult{Arg.getResultType(), ValueKind::Value};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Constant Constant,
-                                               ChildrenT Children) {
-  ConstantData *LiteralValue = Constant.getLiteralValue();
-
-  // FIXME(KZHURAVL): Support ConstantFP?
-  ConstantInt *IntLiteralValue = dyn_cast<ConstantInt>(LiteralValue);
-  if (!IntLiteralValue)
-    return std::nullopt;
-
-  if (isUnsigned(IntLiteralValue)) {
-    emitUnsigned(IntLiteralValue->getZExtValue());
-  } else {
-    emitSigned(IntLiteralValue->getSExtValue());
-  }
-
-  return OpResult{IntLiteralValue->getType(), ValueKind::Value};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::PushLane PushLane,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Referrer ReferrerOp,
-                                               ChildrenT Children) {
-  if (Referrer->isReg() && Referrer->getReg()) {
-    auto DWARFRegister = TRI->getDwarfRegNum(Referrer->getReg(), false);
-    if (DWARFRegister == -1) {
-      return std::nullopt;
-    }
-    emitReg(DWARFRegister);
-    return OpResult{ReferrerOp.getResultType(), ValueKind::LocationDesc};
-  }
-
-  if (Referrer->isImm()) {
-    auto I = Referrer->getImm();
-    if (I >= 0)
-      emitUnsigned(static_cast<uint64_t>(I));
-    else
-      emitSigned(I);
-    return OpResult{ReferrerOp.getResultType(), ValueKind::Value};
-  }
-
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::TypeObject TypeObject,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::AddrOf AddrOf,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Convert Convert,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::ZExt ZExt,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::SExt SExt,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Deref Deref,
-                                               ChildrenT Children) {
-  auto Child = traverse(Children[0].get(), ValueKind::LocationDesc);
-  if (!Child)
-    return std::nullopt;
-
-  // FIXME(KZHURAVL): Support non pointer types?
-  if (!Child->Ty->isPointerTy())
-    return std::nullopt;
-
-  PointerType *PointerResultType = dyn_cast<PointerType>(Child->Ty);
-  assert(PointerResultType && "Expected PointerType, but got something else");
-
-  uint64_t PointerSizeInBits = AP.getDataLayout().getPointerSizeInBits(
-      PointerResultType->getAddressSpace());
-  assert(PointerSizeInBits % 8 == 0 && "Expected multiple of 8");
-
-  uint64_t PointerSizeInBytes = PointerSizeInBits / 8;
-
-  unsigned PointerLLVMAddrSpace = PointerResultType->getAddressSpace();
-  auto PointerDWARFAddrSpace = AP.TM.mapToDWARFAddrSpace(PointerLLVMAddrSpace);
-  if (!PointerDWARFAddrSpace) {
-    LLVM_DEBUG(dbgs() << "Failed to lower DIOpDeref of pointer to addrspace("
-                      << PointerLLVMAddrSpace
-                      << "): no corresponding DWARF addrspace.\n");
-    return std::nullopt;
-  }
-
-  emitDwarfOp(dwarf::DW_OP_deref_size);
-  emitDwarfData1(PointerSizeInBytes);
-  emitDwarfOp(dwarf::DW_OP_constu);
-  emitDwarfUnsigned(*PointerDWARFAddrSpace);
-  emitDwarfOp(dwarf::DW_OP_LLVM_form_aspace_address);
-
-  // FIXME(KZHURAVL): Is the following result type correct?
-  return OpResult{Deref.getResultType(), ValueKind::LocationDesc};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Extend Extend,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Read Read,
-                                               ChildrenT Children) {
-  auto Child = traverse(Children[0].get(), ValueKind::LocationDesc);
-  if (!Child)
-    return std::nullopt;
-  readToValue(Children[0].get());
-  return OpResult{Child->Ty, ValueKind::Value};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Reinterpret Reinterpret,
-                                               ChildrenT Children) {
-  auto Child = traverse(Children[0].get(), ValueKind::LocationDesc);
-  if (!Child)
-    return Child;
-  return OpResult{Reinterpret.getResultType(), Child->VK};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Select Select,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Composite Composite,
-                                               ChildrenT Children) {
-  return std::nullopt;
-}
-
-std::optional<OpResult> DwarfExprAST::traverseMathOp(uint8_t DwarfOp,
-                                                     ChildrenT Children) {
-  auto LHS = traverse(Children[0].get(), ValueKind::Value);
-  if (!LHS)
-    return std::nullopt;
-  auto RHS = traverse(Children[1].get(), ValueKind::Value);
-  if (!RHS)
-    return std::nullopt;
-
-  emitDwarfOp(DwarfOp);
-  return OpResult{LHS->Ty, ValueKind::Value};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::ByteOffset ByteOffset,
-                                               ChildrenT Children) {
-  auto LHS = traverse(Children[0].get(), ValueKind::LocationDesc);
-  if (!LHS)
-    return std::nullopt;
-  auto RHS = traverse(Children[1].get(), ValueKind::Value);
-  if (!RHS)
-    return std::nullopt;
-
-  emitDwarfOp(dwarf::DW_OP_LLVM_offset);
-  return OpResult{ByteOffset.getResultType(), ValueKind::LocationDesc};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::BitOffset BitOffset,
-                                               ChildrenT Children) {
-  auto LHS = traverse(Children[0].get(), ValueKind::LocationDesc);
-  if (!LHS)
-    return std::nullopt;
-  auto RHS = traverse(Children[1].get(), ValueKind::Value);
-  if (!RHS)
-    return std::nullopt;
-
-  emitDwarfOp(dwarf::DW_OP_LLVM_bit_offset);
-  return OpResult{BitOffset.getResultType(), ValueKind::LocationDesc};
-}
-
-std::optional<OpResult> DwarfExprAST::traverse(DIOp::Fragment Fragment,
-                                               ChildrenT Children) {
-  llvm_unreachable("should have dropped fragments by now");
-  return std::nullopt;
-}
-
-void DwarfExprAST::readToValue(Type *Ty) {
-  uint64_t PrimitiveSizeInBits = Ty->getPrimitiveSizeInBits();
-  assert(PrimitiveSizeInBits != 0 && "Expected primitive type");
-
-  uint64_t ByteAlignedPrimitiveSizeInBits = alignTo<8>(PrimitiveSizeInBits);
-  uint64_t PrimitiveSizeInBytes = ByteAlignedPrimitiveSizeInBits / 8;
-  bool NeedsMask = ByteAlignedPrimitiveSizeInBits != PrimitiveSizeInBits;
-
-  emitDwarfOp(dwarf::DW_OP_deref_size);
-  emitDwarfData1(PrimitiveSizeInBytes);
-
-  if (NeedsMask) {
-    uint64_t Mask = (1ULL << PrimitiveSizeInBits) - 1ULL;
-    emitDwarfOp(dwarf::DW_OP_constu);
-    emitDwarfUnsigned(Mask);
-    emitDwarfOp(dwarf::DW_OP_and);
-  }
-}
-
-void DwarfExprAST::readToValue(DwarfExprAST::Node *OpNode) {
-  assert(OpNode->isLowered() && "Expected lowered node");
-  assert(OpNode->getResultType() && "Expected non-null result type");
-  readToValue(OpNode->getResultType());
-}
-
-void DwarfExprAST::emitReg(int32_t DwarfReg, const char *Comment) {
-  assert(DwarfReg >= 0 && "Invalid dwarf register number");
-
-  if (DwarfReg < 32) {
-    emitDwarfOp(dwarf::DW_OP_reg0 + DwarfReg, Comment);
-  } else {
-    emitDwarfOp(dwarf::DW_OP_regx, Comment);
-    emitDwarfUnsigned(DwarfReg);
-  }
-}
-
-void DwarfExprAST::emitSigned(int64_t SignedValue) {
-  emitDwarfOp(dwarf::DW_OP_consts);
-  emitDwarfSigned(SignedValue);
-}
-
-void DwarfExprAST::emitUnsigned(uint64_t UnsignedValue) {
-  if (UnsignedValue < 32) {
-    emitDwarfOp(dwarf::DW_OP_lit0 + UnsignedValue);
-  } else if (UnsignedValue == std::numeric_limits<uint64_t>::max()) {
-    // Only do this for 64-bit values as the DWARF expression stack uses
-    // target-address-size values.
-    emitDwarfOp(dwarf::DW_OP_lit0);
-    emitDwarfOp(dwarf::DW_OP_not);
-  } else {
-    emitDwarfOp(dwarf::DW_OP_constu);
-    emitDwarfUnsigned(UnsignedValue);
-  }
-}
-
-ByteStreamer &DebugLocDwarfExprAST::getActiveStreamer() {
-  return OutBS;
-}
-
-void DebugLocDwarfExprAST::emitDwarfData1(uint8_t Data1Value) {
-  getActiveStreamer().emitInt8(Data1Value, Twine(Data1Value));
-}
-
-void DebugLocDwarfExprAST::emitDwarfOp(uint8_t DwarfOpValue,
-                                       const char *Comment) {
-  if (EmitHeterogeneousDwarfAsUserOps) {
-    if (auto OptUserOp = dwarf::getUserOp(DwarfOpValue)) {
-      getActiveStreamer().emitInt8(
-          dwarf::DW_OP_LLVM_user,
-          dwarf::OperationEncodingString(dwarf::DW_OP_LLVM_user));
-      DwarfOpValue = *OptUserOp;
-    }
-  }
-  getActiveStreamer().emitInt8(
-      DwarfOpValue, Comment ? Twine(Comment) + " " +
-                                  dwarf::OperationEncodingString(DwarfOpValue)
-                            : dwarf::OperationEncodingString(DwarfOpValue));
-}
-
-void DebugLocDwarfExprAST::emitDwarfSigned(int64_t SignedValue) {
-  getActiveStreamer().emitSLEB128(SignedValue, Twine(SignedValue));
-}
-
-void DebugLocDwarfExprAST::emitDwarfUnsigned(uint64_t UnsignedValue) {
-  getActiveStreamer().emitULEB128(UnsignedValue, Twine(UnsignedValue));
-}
-
-void DebugLocDwarfExprAST::emitDwarfAddr(const MCSymbol *Sym) {
-  // FIXME: I'm not sure how to handle IsImplemented when using
-  // BufferByteStreamer; it isn't obvious how to "discard" only the result of
-  // the current expr.
-  IsImplemented = false;
-}
-
-void DebugLocDwarfExprAST::emitDwarfOpAddrx(unsigned Index) {
-  IsImplemented = false;
-}
-
-void DebugLocDwarfExprAST::emitDwarfLabelDelta(const MCSymbol *Hi,
-                                               const MCSymbol *Lo) {
-  IsImplemented = false;
-}
-
-DIELoc &DIEDwarfExprAST::getActiveDIE() {
-  return OutDIE;
-}
-
-void DIEDwarfExprAST::emitDwarfData1(uint8_t Data1Value) {
-  CU.addUInt(getActiveDIE(), dwarf::DW_FORM_data1, Data1Value);
-}
-
-void DIEDwarfExprAST::emitDwarfOp(uint8_t DwarfOpValue, const char *Comment) {
-  if (EmitHeterogeneousDwarfAsUserOps) {
-    if (auto OptUserOp = dwarf::getUserOp(DwarfOpValue)) {
-      CU.addUInt(getActiveDIE(), dwarf::DW_FORM_data1, dwarf::DW_OP_LLVM_user);
-      DwarfOpValue = *OptUserOp;
-    }
-  }
-  CU.addUInt(getActiveDIE(), dwarf::DW_FORM_data1, DwarfOpValue);
-}
-
-void DIEDwarfExprAST::emitDwarfSigned(int64_t SignedValue) {
-  CU.addSInt(getActiveDIE(), dwarf::DW_FORM_sdata, SignedValue);
-}
-
-void DIEDwarfExprAST::emitDwarfUnsigned(uint64_t UnsignedValue) {
-  CU.addUInt(getActiveDIE(), dwarf::DW_FORM_udata, UnsignedValue);
-}
-
-void DIEDwarfExprAST::emitDwarfAddr(const MCSymbol *Sym) {
-  CU.addLabel(getActiveDIE(), dwarf::DW_FORM_addr, Sym);
-}
-
-void DIEDwarfExprAST::emitDwarfOpAddrx(unsigned Index) {
-  bool HasOpAddrx = CU.getDwarfDebug().getDwarfVersion() >= 5;
-  emitDwarfOp(HasOpAddrx ? dwarf::DW_OP_addrx : dwarf::DW_OP_GNU_addr_index);
-  CU.addUInt(getActiveDIE(),
-             HasOpAddrx ? dwarf::DW_FORM_addrx : dwarf::DW_FORM_GNU_addr_index,
-             Index);
-}
-
-void DIEDwarfExprAST::emitDwarfLabelDelta(const MCSymbol *Hi,
-                                          const MCSymbol *Lo) {
-  CU.addLabelDelta(getActiveDIE(), (dwarf::Attribute)0, Hi, Lo);
-}
